<?php
/**
 * SphiderLite Version 2.6.*
 * Sphider is licensed under the GNU GPL.
 *
 * PHP version 5.4
 *
 * @category Sphider_Search
 * @package  Admin
 * @author   R. Prew  <CaptQ@blog.worldspaceflight.com>
 * @license  GNU GPL
 * @link     http://www.sphidersearch.com
 *
 * Original base code by:
 *   Ando Saabas  ando(a t)set.ee
 */

/**
 * Get the file contents
 *
 * @param string $url URL to parse
 *
 * @return array $contents Arry containing page content
 */
function getFileContents($url)
{
    global $user_agent;

    $urlparts = parse_url($url);
    $path = $urlparts['path'];
    $host = $urlparts['host'];
    if ($urlparts['query'] != "") {
        $path .= "?".$urlparts['query'];
    }
    if (isset($urlparts['port'])) {
        $port = (int) $urlparts['port'];
    } else {
        if ($urlparts['scheme'] == "http") {
            $port = 80;
        } else {
            if ($urlparts['scheme'] == "https") {
                $port = 443;
            }
        }
    }

    if ($port == 80) {
        $portq = "";
    } else {
        $portq = ":$port";
    }

    $all = "*/*";

    $request = "GET $path HTTP/1.0\r\nHost: $host$portq\r\nAccept: "
      ."$all\r\nUser-Agent: $user_agent\r\n\r\n";

    $fsocket_timeout = 30;
    if (mb_substr($url ?? '', 0, 5, 'utf-8') == "https") {
        $target = "ssl://".$host;
    } else {
        $target = $host;
    }

    $errno = 0;
    $errstr = "";
    $fp = @ fsockopen($target, $port, $errno, $errstr, $fsocket_timeout);

    print $errstr;
    if (!$fp) {
        $contents['state'] = "NOHOST";
        printConnectErrorReport($errstr);
        return $contents;
    } else {
        if (!fputs($fp, $request)) {
            $contents['state'] = "Cannot send request";
            return $contents;
        }
        $data = null;
        socket_set_timeout($fp, $fsocket_timeout);
        do {
            $status = socket_get_status($fp);
            $data .= fgets($fp, 8192);
        } while (!feof($fp) && !$status['timed_out']);

        fclose($fp);
        if ($status['timed_out'] == 1) {
            $contents['state'] = "timeout";
        } else {
            $contents['state'] = "ok"; 
        }
        $contents['file'] = mb_substr(
            $data ?? '', mb_strpos($data, "\r\n\r\n", 0, 'utf-8') + 4, null, 'utf-8'
        );
    }
    return $contents;
}


/**
 * Check if file is available and in readable form
 *
 * @param string $url The url to checkl
 *
 * @return array $status Array of status information
 */
function urlStatus($url)
{
    global $user_agent, $index_pdf, $index_doc, $index_xls, $index_ppt;

    $urlparts = parse_url($url);
    $path = $urlparts['path'];
    $host = $urlparts['host'];
    if (isset($urlparts['query'])) {
        $path .= "?".$urlparts['query'];
    }
    if (isset($urlparts['port'])) {
        $port = (int) $urlparts['port'];
    } else {
        if ($urlparts['scheme'] == "http") {
             $port = 80;
        } else {
            if ($urlparts['scheme'] == "https") {
                $port = 443;
            }
        }
    }
    if ($port == 80) {
        $portq = "";
    } else {
        $portq = ":$port";
    }

    $all = "*/*"; //just to prevent "comment effect" in get accept
    $request = "HEAD $path HTTP/1.1\r\nHost: $host$portq\r\nAccept: "
      ."$all\r\nUser-Agent: $user_agent\r\n\r\n";

    if (mb_substr($url ?? '', 0, 5, 'utf-8') == "https") {
        $target = "ssl://".$host;
    } else {
        $target = $host;
    }

    $fsocket_timeout = 30;
    $errno = 0;
    $errstr = "";
    $fp = fsockopen($target, $port, $errno, $errstr, $fsocket_timeout);
    print $errstr;
    $linkstate = "ok";
    if (!$fp) {
        $status['state'] = "NOHOST";
    } else {
        socket_set_timeout($fp, 30);
        fputs($fp, $request);
        $answer = fgets($fp, 4096);
        if (is_null($answer)) {
            $answer = "";
        }
        $regs = Array ();
        if (preg_match("#HTTP/[0-9.]+ (([0-9])[0-9]{2})#", $answer, $regs)) {
            $httpcode = $regs[2];
            $full_httpcode = $regs[1];
            if ($httpcode <> 2 && $httpcode <> 3) {
                $status['state'] = "Unreachable: http $full_httpcode";
                $linkstate = "Unreachable";
            }
        }
        if ($linkstate <> "Unreachable") {
            while ($answer) {
                $answer = fgets($fp, 4096);

                if (preg_match("/Location: *([^\n\r ]+)/", $answer, $regs)
                    && $httpcode == 3 && $full_httpcode != 302
                ) {
                    $status['path'] = $regs[1];
                    $status['state'] = "Relocation: http $full_httpcode";
                    fclose($fp);
                    return $status;
                }

                if (preg_match(
                    "/Last-Modified: *([a-z0-9,: ]+)/i", $answer, $regs
                )
                ) {
                    $status['date'] = $regs[1];
                }

                if (preg_match("/Content-Type:/i", $answer)) {
                    $content = $answer;
                    $answer = '';
                    break;
                }
            }
            $socket_status = socket_get_status($fp);
            if (is_null($content)) {
                $content = "";
            }
            $docx = 'application/vnd.openxmlformats-officedocument';
            $docx = $docx.'.wordprocessingml.document';
            if (preg_match("/Content-Type: *([a-z\/.-]*)/i", $content, $regs)) {
                if ($regs[1] == 'text/html' || $regs[1] == 'text/'
                    || $regs[1] == 'text/plain'
                ) {
                    $status['content'] = 'text';
                    $status['state'] = 'ok';
                } else if ($regs[1] == 'application/pdf' && $index_pdf == 1) {
                    $status['content'] = 'pdf';
                    $status['state'] = 'ok';
                } else if (($regs[1] == 'application/msword'
                    || $regs[1] == 'application/vnd.ms-word') && $index_doc == 1
                ) {
                    $status['content'] = 'doc';
                    $status['state'] = 'ok';
                } else if (($regs[1] == 'application/excel'
                    || $regs[1] == 'application/vnd.ms-excel') && $index_xls == 1
                ) {
                    $status['content'] = 'xls';
                    $status['state'] = 'ok';
                } else if (($regs[1] == 'application/mspowerpoint'
                    || $regs[1] == 'application/vnd.ms-powerpoint')
                    && $index_ppt == 1
                ) {
                    $status['content'] = 'ppt';
                    $status['state'] = 'ok';
                } else if (($regs[1] == $docx)
                    && $index_doc == 1
                ) {
                    $status['content'] = 'docx';
                    $status['state'] = 'ok';
                } else if (($regs[1] == 'application/vnd.oasis.opendocument.text')
                    && $index_doc == 1
                ) {
                    $status['content'] = 'odt';
                    $status['state'] = 'ok';
                } else {
                    $status['state'] = "Not text or html";
                }
            } else {
                if ($socket_status['timed_out'] == 1) {
                    $status['state'] = "Timed out (no reply from server)";
                } else {
                    $status['state'] = "Not text or html";
                }
            }
        }
    }
    fclose($fp);
    return $status;
}


/**
 * Read robots.txt file in the server, to find any disallowed files/folders
 *
 * @param string $url The url to check for a robots.txt file
 *
 * @return array $omit An array of forbidden and permitted urls
 */
function checkRobotTxt($url)
{
    global $user_agent;
    
    $urlparts = parse_url($url);
    $url = $urlparts['scheme']."://".$urlparts['host']."/robots.txt";

    $url_status = urlStatus($url);
    $result = array();
    $final = array (
      'allow' =>array(),
      'deny' =>array()
    );
    $spider_agent = array (
      'allow' =>array(),
      'deny' =>array()
    );
    $star_agent = array (
      'allow' =>array(),
      'deny' =>array()
    );

    if ($url_status['state'] == "ok") {
        $robot = file($url);
        if (!$robot) {
            $contents = getFileContents($url);
            $file = $contents['file'];
            $robot = explode("\n", $file);
        }

        $regs = Array ();
        $this_agent= "";
        foreach ($robot as $id => $line) {
            if (preg_match("/^[uU]ser-[aA]gent: *([^#]+) */", $line ?? '', $regs)) {
                $this_agent = trim($regs[1]);
                // We only care about Sphider and * user agents. All others are
                // irrelevent to our needs. Check = 1 is relevent, 0 is not
                if ($this_agent == '*' || $this_agent == $user_agent) {
                    $check = 1; 
                } else {
                    $check = 0; 
                }
            }

            // Check for Disallows
            if (preg_match(
                "/^[dD]isallow: *([^#]+)/", $line ?? '', $regs
            ) && $check == 1
            ) {
                $disallow_str = preg_replace("/[\n ]+/", "", $regs[1] ?? '');
                if (trim($disallow_str) != "") {
                    if ($this_agent == '*') {
                        $star_agent['deny'][] = $disallow_str;
                    } else {
                        $spider_agent['deny'][] = $disallow_str;
                    }
                }
            } 
                                   
            // Check for Allows
            if (preg_match(
                "/^[aA]llow: *([^#]+)/", $line ?? '', $regs
            ) && $check == 1
            ) {
                $allow_str = preg_replace("/[\n ]+/", "", $regs[1] ?? '');
                if (trim($allow_str) != "") {
                    if ($this_agent == '*') {
                        $star_agent['allow'][] = $allow_str;
                    } else {
                        $spider_agent['allow'][] = $allow_str;
                    }
                }
            }
        }           
    } 
         
         
    // Resolve conflict in the Agent * array
    $result = array_intersect($star_agent['deny'], $star_agent['allow']);
    foreach ($result as $id => $value) {
        unset($star_agent['deny'][$id]);
    }
    sort($star_agent['deny']);

    // Resolve conflict in the Agent Sphider array
    $result = array_intersect($spider_agent['deny'], $spider_agent['allow']);
    foreach ($result as $id => $value) {
        unset($spider_agent['deny'][$id]);
    }
    sort($spider_agent['deny']);

    // Resolve conflict between Sphider allow and * deny
    $result = array_intersect($star_agent['deny'], $spider_agent['allow']);
    foreach ($result as $id => $value) {
        unset($star_agent['deny'][$id]);
    }
    sort($star_agent['deny']);

    // Resolve conflict between Sphider deny and * allow
    $result = array_intersect($star_agent['allow'], $spider_agent['deny']);
    foreach ($result as $id => $value) {
        unset($star_agent['allow'][$id]);
    }
    sort($star_agent['allow']);

    // Handle special case of Sphider allow '/'
    foreach ($spider_agent['allow'] as $id => $value) {
        if ($value === "/") {
            foreach ($star_agent['deny'] as $j => $k) {
                unset($star_agent['deny'][$j]);
            }
            sort($star_agent['deny']);
        }
    }

    // Handle special case of Sphider deny'/'
    foreach ($spider_agent['deny'] as $id => $value) {
        if ($value === "/") {
            foreach ($star_agent['allow'] as $j => $k) {
                unset($star_agent['allow'][$j]);
            }
            sort($star_agent['allow']);
        }
    }

    // Combine the agent Sphider and agent * arrays
    foreach ($spider_agent['allow'] as $id => $value) {
        $final['allow'][] = $value;
    }
    foreach ($star_agent['allow'] as $id => $value) {
        $final['allow'][] = $value;
    }
    foreach ($spider_agent['deny'] as $id => $value) {
        $final['deny'][] = $value;
    }
    foreach ($star_agent['deny'] as $id => $value) {
        $final['deny'][] = $value;
    }
    
    // Remove any duplicate entries remaining
    array_unique($final['allow']);
    array_unique($final['deny']);

    return $final;
}


/**
 * Read the sitemap.xml file on the server
 *
 * @param string $input_file Sitemap file name
 *
 * @return array $links Array of links found in sitemap.xml
 */
function getSiteMap($input_file)
{
    $links = '';
    $sitemap = simplexml_load_file($input_file);
    if ($sitemap != '') {
        $links = array ();
        foreach ($sitemap as $url) {
            // For some reason, wlwmanifest.xml interfers with the recursion
            // Therefore, let's ignore it
            if (preg_match("/wlwmanifest\.xml$/i", $url->loc ?? '')) {
                continue;
            }
            if (preg_match("/\.xml.gz$/i", $url->loc ?? '')) {
                $submap_gz = $url->loc;
                foreach ($submap_gz as $input3) {
                    $zd = gzopen($input3, "r");
                    $contents = gzread($zd, 100000);
                    gzclose($zd);
                    $sitemap_gz = simplexml_load_string($contents);
                    if ($sitemap_gz != '') {
                        foreach ($sitemap_gz as $url_gz) {
                            $links[] = ($url_gz->loc);
                        }
                    }
                }
            }
            if (preg_match("/\.xml$/i", $url->loc ?? '')) {
                $submap = $url->loc;
                foreach ($submap as $input2) {
                    $sitemap2 = simplexml_load_file($input2);
                    if ($sitemap2 != '') {
                        foreach ($sitemap2 as $url2) {
                            $links[] = ($url2->loc);
                        }
                    }
                }
            } else {
                if (!preg_match("/\.gz$/i", $url->loc)) {
                    $links[] =($url->loc);
                }
            }
        }
            $links = explode(",", (implode(",", $links)));
    }
    return $links;
}


/**
 * Remove the file part from an url (to build an url from an url
 * and given relative path)
 *
 * @param string $url Url to be removed
 *
 * @return string $url Updated url string
 */
function removeFileFromUrl($url)
{
    $url_parts = parse_url($url);
    $path = $url_parts['path'];
    if (is_null($path)) {
        $path = "";
    }

    $regs = Array ();
    if (preg_match('/([^\/]+)$/i', $path, $regs)) {
        $file = $regs[1];
        $check = $file.'$';
        $path = preg_replace("/$check"."/i", "", $path);
    }

    if ($url_parts['port'] == 80 || $url_parts['port'] == "") {
        $portq = "";
    } else {
        $portq = ":".$url_parts['port'];
    }

    $url = $url_parts['scheme']."://".$url_parts['host'].$portq.$path;
    return $url;
}


/**
 * Extract links from html
 *
 * @param string $file             File containing page contents
 * @param string $url              Url of the file source
 * @param int    $can_leave_domain Indicates if spider may leave the domain
 * @param string $base             Base url
 *
 * @return array $links Array of links found
 */
function getLinks($file, $url, $can_leave_domain, $base)
{
    $chunklist = array ();
    // The base URL comes from either the meta tag or the current URL.
    if (!empty($base)) {
        $url = $base;
    }

    $links = array ();
    $regs = Array ();
    $checked_urls = Array();

    preg_match_all(
        "/[^\?]href\s*=\s*[\'\"]?([+:%\/\?~=&;\\\(\),._a-zA-Z0-9-]*)"
        ."(#[.a-zA-Z0-9-]*)?[\'\" ]?(\s*rel\s*=\s*[\'\"]?(nofollow)[\'\"]?)?/i",
        $file, $regs, PREG_SET_ORDER
    );
    foreach ($regs as $val) {
        //if nofollow is not set
        if ($checked_urls[$val[1]]!=1 && !isset($val[4])) {
            if (($a = urlPurify($val[1], $url, $can_leave_domain)) != '') {
                $links[] = $a;
            }
            $checked_urls[$val[1]] = 1;
        }
    }
    preg_match_all(
        "/(frame[^>]*src[[:blank:]]*)=[[:blank:]]*[\'\"]?(([[a-z]{3,5}:\/\/"
        ."(([.a-zA-Z0-9-])+(:[0-9]+)*))*([+:%\/?=&;\\\(\),._ a-zA-Z0-9-]*))"
        ."(#[.a-zA-Z0-9-]*)?[\'\" ]?/i", $file, $regs, PREG_SET_ORDER
    );
    foreach ($regs as $val) {
        if ($checked_urls[$val[1]]!=1 && !isset($val[4])) {
            // if nofollow is not set
            if (($a = urlPurify($val[1], $url, $can_leave_domain)) != '') {
                $links[] = $a;
            }
            $checked_urls[$val[1]] = 1;
        }
    }
    preg_match_all(
        "/(iframe[^>]*src[[:blank:]]*)=[[:blank:]]*[\'\"]?(([[a-z]{3,5}:\/\/"
        ."(([.a-zA-Z0-9-])+(:[0-9]+)*))*([+:%\/?=&;\\\(\),._ a-zA-Z0-9-]*))"
        ."(#[.a-zA-Z0-9-]*)?[\'\" ]?/i", $file, $regs, PREG_SET_ORDER
    );
    foreach ($regs as $val) {
        if ($checked_urls[$val[2]]!=1 ) {
            // if nofollow is not set
            if (($a = urlPurify($val[2], $url, $can_leave_domain)) != '') {
                $links[] = $a;
            }
            $checked_urls[$val[2]] = 1;
        }
    }
    preg_match_all(
        "/(window[.]location)[[:blank:]]*=[[:blank:]]*[\'\"]?(([[a-z]{3,5}:\/\/"
        ."(([.a-zA-Z0-9-])+(:[0-9]+)*))*([+:%\/?=&;\\\(\),._ a-zA-Z0-9-]*))"
        ."(#[.a-zA-Z0-9-]*)?[\'\" ]?/i", $file, $regs, PREG_SET_ORDER
    );
    foreach ($regs as $val) {
        if ($checked_urls[$val[1]]!=1 && !isset($val[4])) {
            // if nofollow is not set
            if (($a = urlPurify($val[1], $url, $can_leave_domain)) != '') {
                $links[] = $a;
            }
            $checked_urls[$val[1]] = 1;
        }
    }
    preg_match_all(
        "/(http-equiv=['\"]refresh['\"] *content=['\"][0-9]+;url)[[:blank:]]*=["
        ."[:blank:]]*[\'\"]?(([[a-z]{3,5}:\/\/(([.a-zA-Z0-9-])+(:[0-9]+)*))*"
        ."([+:%\/?=&;\\\(\),._ a-zA-Z0-9-]*))(#[.a-zA-Z0-9-]*)?[\'\" ]?/i",
        $file, $regs, PREG_SET_ORDER
    );
    foreach ($regs as $val) {
        if ($checked_urls[$val[1]]!=1 && !isset($val[4])) {
            // if nofollow is not set
            if (($a = urlPurify($val[1], $url, $can_leave_domain)) != '') {
                $links[] = $a;
            }
            $checked_urls[$val[1]] = 1;
        }
    }
    preg_match_all(
        "/(window[.]open[[:blank:]]*[(])[[:blank:]]*[\'\"]?(([[a-z]{3,5}:\/\/"
        ."(([.a-zA-Z0-9-])+(:[0-9]+)*))*([+:%\/?=&;\\\(\),._ a-zA-Z0-9-]*))"
        ."(#[.a-zA-Z0-9-]*)?[\'\" ]?/i", $file, $regs, PREG_SET_ORDER
    );
    foreach ($regs as $val) {
        if ($checked_urls[$val[1]]!=1 && !isset($val[4])) {
            // if nofollow is not set
            if (($a = urlPurify($val[1], $url, $can_leave_domain)) != '') {
                $links[] = $a;
            }
            $checked_urls[$val[1]] = 1;
        }
    }

    return $links;
}


/**
 * Function to build a unique word array from the text of a webpage,
 * together with the count of each word
 *
 * @param array $arr Array of page content
 *
 * @return array $newarr Array with redundencies removed
 */
function uniqueArray($arr)
{
    global $min_word_length;
    global $common, $ctlang, $pglang;
    global $word_upper_bound;
    global $index_numbers, $index_decimals, $stem_words, $stem;

    if ($stem_words == 1) {
        $newarr = Array();
        foreach ($arr as $val) {
            $newarr[] = $stem->stem($val);
        }
        $arr = $newarr;
    }
    sort($arr);
    reset($arr);
    $newarr = array ();

    $i = 0;
    $counter = 1;
    $element = current($arr);

    if ($index_numbers == 1) {
        $pattern = "/[\pL0-9]+/";
    } else {
        $pattern = "/[\pL]+/";
    }

    if (!is_null($pglang)) {
        $uselang = $pglang;
    } else {
        $uselang = $ctlang;
    }

    $regs = Array ();
    for ($n = 0; $n < sizeof($arr); $n ++) {
        // check if word is long enough, contains alphabetic characters and
        // is not a common word to eliminate/count multiple instance of words
        $next_in_arr = next($arr);
        if ($next_in_arr != $element) {
            if (mb_strlen($element, 'utf-8') >= $min_word_length
                && preg_match($pattern, $element ?? '')
                && (@ $common[$uselang][$element] <> 1)
            ) {
                if (preg_match("/^(-|\\\')(.*)/", $element ?? '', $regs)) {
                    $element = $regs[2]; 
                }

                if (preg_match("/(.*)(\\\'|-)$/", $element ?? '', $regs)) {
                    $element = $regs[1]; 
                }
                $newarr[$i][1] = $element;
                $newarr[$i][2] = $counter;
                $element = current($arr);
                $i ++;
                $counter = 1;
            } else {
                 $element = $next_in_arr;
            }
        } else {
            if ($counter < $word_upper_bound) {
                $counter ++; 
            }
        }
    }
    return $newarr;
}


/**
 * Checks if url is legal, relative to the main url.
 *
 * @param string $url              Url to be checked
 * @param string $parent_url       Url containing the link
 * @param int    $can_leave_domain Indicates if spider can leave the domain
 *
 * @return string $url Purified url, empty if not to be indexed
 */
function urlPurify($url, $parent_url, $can_leave_domain)
{
    global $ext, $mainurl, $apache_indexes, $strip_sessids;

    if (is_null($url)) {
        $url = "";
    }
    if (substr($url, 0, 2) == "./") {
        $url = substr($url, 2);
    }
    $urlparts = parse_url($url);

    $main_url_parts = parse_url($mainurl);
    if ($urlparts['host'] != "" && $urlparts['host'] != $main_url_parts['host']
        && $can_leave_domain != 1
    ) {
        return '';
    }

    reset($ext);
    foreach ($ext as $id => $excl) {
        if (preg_match("/\.$excl$/i", $url)) {
            return ''; 
        }
    }

    if (mb_substr($url ?? '', -1, null, 'utf-8') == '\\') {
        return '';
    }

    if (isset($urlparts['query'])) {
        if ($apache_indexes[$urlparts['query']]) {
            return '';
        }
    }

    if (preg_match("/[\/]?mailto:|[\/]?javascript:|[\/]?news:/i", $url)) {
        return '';
    }

    if (isset($urlparts['scheme'])) {
        $scheme = $urlparts['scheme'];
    } else {
        $scheme ="";
    }

    //only http and https links are followed
    if (!($scheme == 'http' || $scheme == '' || $scheme == 'https')) {
        return '';
    }

    //parent url might be used to build an url from relative path
    $parent_url = removeFileFromUrl($parent_url);
    $parent_url_parts = parse_url($parent_url);

    if (mb_substr($url ?? '', 0, 1, 'utf-8') == '/') {
        $url = $parent_url_parts['scheme']."://".$parent_url_parts['host'].$url;
    } else if (!isset($urlparts['scheme'])) {
        $url = $parent_url.$url;
    }

    $url_parts = parse_url($url);

    $urlpath = $url_parts['path'];
    if (is_null($urlpath)) {
        $urlpath = "";
    }

    $regs = Array ();
    
    while (preg_match("/[^\/]*\/[.]{2}\//", $urlpath, $regs)) {
        $urlpath = str_replace($regs[0], "", $urlpath);
    }

    //remove relative path instructions like ../ etc
    $urlpath = preg_replace("/\/+/", "/", $urlpath);
    $urlpath = preg_replace("/[^\/]*\/[.]{2}/", "",  $urlpath);
    $urlpath = str_replace("./", "", $urlpath);
    $query = "";
    if (isset($url_parts['query'])) {
        $query = "?".$url_parts['query'];
    }

    if ($main_url_parts['port'] == 80 || $url_parts['port'] == "") {
        $portq = "";
    } else {
        $portq = ":".$main_url_parts['port'];
    }
    $url = $url_parts['scheme']."://".$url_parts['host'].$portq.$urlpath.$query;

    //if we index sub-domains
    if ($can_leave_domain == 1) {
        return $url;
    }

    $mainurl = removeFileFromUrl($mainurl);
    
    if ($strip_sessids == 1) {
        $url = removeSessId($url);
    }

    //only urls in staying in the starting domain/directory are followed
    $url = convertUrl($url);
    if (strstr($url, $mainurl) == false) {
        return '';
    } else {
        return $url; 
    }
}


/**
 * Save keywords to the DB
 *
 * @param array $wordarray Array of words to be stored
 * @param int   $link_id   Numberic id of the page containing the word
 * @param int   $domain    Nurmeric id of the domain
 *
 * @return void
 */
function saveKeywords($wordarray, $link_id, $domain)
{
    global $mysql_table_prefix, $all_keywords, $db, $charset;
    global $min_word_length;

    foreach ($wordarray as $key =>$thisword) {
        $word = $thisword[1];
        if ($charset != "" && $charset != "UTF-8") {
            $word = iconv($charset, "UTF-8//IGNORE", $word);
        } else if ($charset == "UTF-8") {
            $word = iconv("UTF-8", "UTF-8//IGNORE", $word);
        }
        $wordmd5 = mb_substr(md5($word) ?? '', 0, 1, 'utf-8');
        $weight = $thisword[2];
        if (mb_strlen($word, 'utf-8') < $min_word_length) {
            continue;
        }
        if (mb_strlen($word, 'utf-8')<= 30) {
            $keyword_id = $all_keywords[$word];
            if ($keyword_id  == "") {
                $stmt = $db->prepare(
                    "INSERT INTO "
                    .$mysql_table_prefix."keywords (keyword) VALUES ( ? )"
                );
                if ($stmt) {
                    $stmt->bind_param("s", $word);
                    $t_up = 1;  // Set thumbs up
                    try {
                        $stmt->execute();
                    } catch (Exception $e) {
                        $t_up = 0;  // Thumbs down! Error, possibly a duplicate key
                        if ($stmt->errno == 1062) {  // Yup, a duplicate key
                            $stmt->close();
                            $stmt2 = $db->prepare(
                                "SELECT keyword_id FROM "
                                .$mysql_table_prefix."keywords WHERE keyword = ? "
                            );
                            if ($stmt2) {
                                $stmt2->bind_param("s", $word);
                                $stmt2->execute() or die(
                                    "Execution failed: "
                                    .$stmt2->error
                                );
                                $stmt2->bind_result($keyword_id);
                                $stmt2->fetch();
                                $stmt2->close();
                            } else {
                                trigger_error(
                                    'Statement failed : '
                                    .$db->error, E_USER_ERROR
                                );
                            }
                        } else {
                            // Other kind of error
                            die("Execution failed: ".$stmt->error);
                        }
                    }
                    if ($t_up == 1) {  // Still thumbs up. proceed with insert
                        $keyword_id = $stmt->insert_id;
                        $all_keywords[$word] = $keyword_id;
                        $stmt->close();
                    }
                } else {
                    trigger_error(
                        'Statement failed : '
                        .$db->error, E_USER_ERROR
                    );
                }
            }
            $inserts[$wordmd5] .= ",('".$link_id."', '".$keyword_id."', '"
              .$weight."', '".$db->real_escape_string($domain)."')";
        }
    }

    $query = "SET FOREIGN_KEY_CHECKS=0";
    $stmt = $db->prepare($query);
    if ($stmt) {
        $stmt->execute() or die("Execution failed: ".$stmt->error);
        $stmt->close();
    } else {
        trigger_error('Statement failed : '.$db->error, E_USER_ERROR);
    }
    for ($i=0;$i<=15; $i++) {
        $char = dechex($i);
        $values= mb_substr($inserts[$char] ?? '', 1, null, 'utf-8');
        if ($values!="") {
            $query = "INSERT INTO ".$mysql_table_prefix."link_keyword".$char
              ." (link_id, keyword_id, weight, domain) VALUES ".$values;
            $stmt = $db->prepare($query);
            if ($stmt) {
                try {
                    $stmt->execute();
                } catch (Exception $e) {
                    // Less stringent than die due to possible errors in
                    // pages being indexed
                    trigger_error(
                        "Execution failed: "
                        .$db->error, E_USER_WARNING
                    );
                }
                $stmt->close();
            } else {
                trigger_error('Statement failed : '.$db->error, E_USER_ERROR);
            }
        }
    }
    $query = "SET FOREIGN_KEY_CHECKS=1";
    $stmt = $db->prepare($query);
    if ($stmt) {
        $stmt->execute() or die("Execution failed: ".$stmt->error);
        $stmt->close();
    } else {
        trigger_error('Statement failed : '.$db->error, E_USER_ERROR);
    }
}


/**
 * Get header data from a file
 *
 * @param string $file File to be parsed
 *
 * @return array $data Array of header data
 */
function getHeadData($file)
{
    $headdata = "";
    
    if (is_null($file)) {
        $file = "";
    }

    preg_match("@<head[^>]*>(.*?)<\/head>@si", $file, $regs);

    $headdata = $regs[1];
    if (is_null($headdata)) {
        $headdata = "";
    }

    $description = "";
    $robots = "";
    $keywords = "";
    $base = "";
    $res = Array ();
    if ($headdata != "") {
        preg_match(
            "/<meta +name *=[\"']?robots[\"']? *content=[\"']?([^<>'\"]+)[\"']?/i",
            $headdata, $res
        );
        if (isset($res)) {
            $robots = $res[1];
        }

        // The first test checks if an apostrophe is used as a delimiter
        // If not, the second test rules
        preg_match(
            "/<meta +name *='?description'? *content='?"
            ."([^<>'\"]+)'?/i", $headdata, $res
        );
        if (isset($res) && $res[1] == "") {
            preg_match(
                "/<meta +name *=[\"']?description[\"']? *content=[\"']?"
                ."([^<>\"]+)[\"']?/i", $headdata, $res
            );    
        }
        if (isset($res)) {
            $description = $res[1];
        }

        preg_match(
            "/<meta +name *=[\"']?keywords[\"']? *content=[\"']?([^<>'\"]+)[\"']?/i",
            $headdata, $res
        );
        if (isset($res)) {
            $keywords = $res[1];
        }
        // e.g. <base href="http://www.consil.co.uk/index.php" />
        preg_match("/<base +href *= *[\"']?([^<>'\"]+)[\"']?/i", $headdata, $res);
        if (isset($res)) {
            $base = $res[1];
        }
        $keywords = preg_replace("/[, ]+/", " ", $keywords ?? '');
        $robots = explode(",", mb_strtolower($robots ?? '', 'utf-8'));
        $nofollow = 0;
        $noindex = 0;
        foreach ($robots as $x) {
            if (trim($x) == "noindex") {
                $noindex = 1;
            }
            if (trim($x) == "nofollow") {
                $nofollow = 1;
            }
        }
        $data['description'] = $description;
        $data['keywords'] = $keywords;
        $data['nofollow'] = $nofollow;
        $data['noindex'] = $noindex;
        $data['base'] = $base;
    }
    return $data;
}


/**
 * Clean the file and organize its content
 *
 * @param string $file    File to be cleaned and organized
 * @param string $url     Url of the file source
 * @param int    $type    Type of file being indexed
 * @param string $charset Character set of file
 *
 * @return array $data Array of page information
 */
function cleanFile($file, $url, $type, $charset)
{
    global $entities, $index_host, $index_decimals, $index_meta_keywords;
    global $dec_sep;

    if (is_null($file)) {
        $file = "";
    }
    $urlparts = parse_url($url);
    $host = $urlparts['host'];
    //remove filename from path
    $path = preg_replace('/([^\/]+)$/i', "", $urlparts['path'] ?? '');
    
    // Convert to UTF-8 if not already
    if ($charset != "" && $charset != "UTF-8") {
        $file = iconv($charset, "UTF-8//IGNORE", $file);
    } else if ($charset == "UTF-8") {
        // Remove illegal utf-8 characters
        $file = iconv("UTF-8", "UTF-8//IGNORE", $file);
    } 
    
    $file = preg_replace("/<link rel[^<>]*>/i", " ", $file);
    $file = preg_replace(
        "@<!--sphider_noindex-->.*?<!--\/sphider_noindex-->@si", " ", $file
    );
    $file = preg_replace("@<!--.*?-->@si", " ", $file);
    $file = preg_replace("@<script[^>]*?>.*?</script>@si", " ", $file);
    $headdata = getHeadData($file);
    $regs = Array ();
    if (preg_match("@<title.*?>(.*?)<\/title.*?>@si", $file, $regs)) {
        $title = trim($regs[1]);
        $file = str_replace($regs[0], "", $file);
    } else if ($type == 'pdf' || $type == 'doc') {
        //the title of a non-html file is its first few words
        $title = mb_substr(
            $file ?? '', 0, mb_strrpos(
                mb_substr($file ?? '', 0, 40, 'utf-8'), " ", 0, 'utf-8'
            ), 'utf-8'
        );
    }
    $file = preg_replace("@<style[^>]*>.*?<\/style>@si", " ", $file);
    //create spaces between tags, so that removing tags doesnt concatenate strings
    $file = preg_replace("/<[\w ]+>/", "\\0 ", $file);
    $file = preg_replace("/<\/[\w ]+>/", "\\0 ", $file);
    $file = strip_tags($file);
    $file = preg_replace("/&nbsp;/", " ", $file);

    $fulltext = $file;
    if (is_null($fulltext)) {
        $fulltext = "";
    }
    // Condense multiple whitespace
    $fulltext = preg_replace("/\s+/", " ", $fulltext);
    $file .= " ".$title;
    if ($index_host == 1) {
        $file = $file." ".$host." ".$path;
    }
    if ($index_meta_keywords == 1) {
        $file = $file." ".$headdata['keywords'];
    }

    //replace codes with ascii chars
    $file = preg_replace_callback(
        '~&#x([0-9a-f]+);~i',
        function ($m) {
            chr(hexdec($m[1]));
        }, $file
    );
    $file = preg_replace_callback(
        '~&#([0-9]+);~',
        function ($m) {
            chr($m[1]);
        }, $file
    );
    // Remove unicode replacement character arising from 
    // improper latin-1 to utf-8 conversion
    $file = preg_replace("/�/", " ", $file ?? '');
    $file = mb_strtolower($file ?? '', 'utf-8');
    foreach ($entities as $key => $char) {
        $file = preg_replace("/".$key."/i", $char, $file);
    }
    $file = preg_replace("/&[a-z]{1,6};/", " ", $file ?? '');
    // Not sure why, but the following characters MUST be escaped
    // to be safely removed!
    // £ ´ … caused unicode corruption when removed in unescaped fashion
    $file = preg_replace("/\£\´\…/", " ", $file ?? ''); 

    // It is desirable to remove the thousands separator from numbers
    // Special case to remove space thousands separators
    if (preg_match_all("/[0-9] [0-9]/", $file ?? '', $matches)) {
        foreach ($matches[0] as $pat) {
            $file = str_replace($pat, $pat[0].$pat[2], $file ?? '');
        }
    }
    // Decimal separator is a decimal period
    if ($dec_sep === ".") {
        // Remove comma thousands separators
        if (preg_match_all("/[0-9],[0-9]/", $file ?? '', $matches)) {
            foreach ($matches[0] as $pat) {
                $file = str_replace($pat, $pat[0].$pat[2], $file ?? '');
            }
        }
        // Decimal separator is a decimal comma
    } else {
        // Remove period thousands separators
        if (preg_match_all("/[0-9].[0-9]/", $file ?? '', $matches)) {
            foreach ($matches[0] as $pat) {
                $file = str_replace($pat, $pat[0].$pat[2], $file ?? '');
            }
        }
    }

    if ($index_decimals == 0) {
        $file = preg_replace(
            "/[\*\^\+\?\\\.\[\]\$\|\{\)\(\}~!\"\/@#%&=`;><:,“]+/", " ", $file ?? ''
        );  // Removed £ ´ … due to unicode corruption
    } elseif ($dec_sep === ".") {
        // Period or full stop as decimal separator
        // Replace period with space if not surrounded by integers,
        // which will allow indexing decimal numbers
        $file = preg_replace("/(?<![0-9])\.|\.(?![0-9])/", " ", $file ?? '');
        $file = preg_replace(
            "/[\*\^\+\?\\\[\]\$\|\{\)\(\}~!\"\/@#%&=`;><:,“]+/", " ", $file ?? ''
        ); // Removed £ ´ … due to unicode corruption
    } else {
        // Comma as decimal separator
        // Replace comma with space if not surrounded by integers,
        // which will allow indexing decimal numbers
        $file = preg_replace("/(?<![0-9]),|,(?![0-9])/", " ", $file ?? '');
        $file = preg_replace(
            "/[\*\^\+\?\\\.\[\]\$\|\{\)\(\}~!\"\/@#%&=`;><:“]+/", " ", $file ?? ''
        ); // Removed £ ´ … due to unicode corruption
    }
    // Replace apostrophe, hyphen if preceded or followed by white space
    $file = preg_replace("/\s['-]+|['-]+\s/", " ", $file);
    $file = preg_replace("/\s+/", " ", $file);
    $data['fulltext'] = $fulltext;
    $data['content'] = $file;
    $data['title'] = $title;
    $data['description'] = $headdata['description'];
    $data['keywords'] = $headdata['keywords'];
    $data['host'] = $host;
    $data['path'] = $path;
    $data['nofollow'] = $headdata['nofollow'];
    $data['noindex'] = $headdata['noindex'];
    $data['base'] = $headdata['base'];

    return $data;

}


/**
 * Calculate keyword weighting
 *
 * @param array  $wordarray Array of words in page content
 * @param string $title     Title of the page
 * @param string $host      Host portion of the url
 * @param string $path      Path portion of the url
 * @param array  $keywords  Array of page information
 *
 * @return arry $wordarray Array of words/weights
 */
function calcWeights($wordarray, $title, $host, $path, $keywords)
{
    global $index_host, $index_meta_keywords;

    if (is_null($host)) {
        $host = "";
    }
    if (is_null($path)) {
        $path = "";
    }
    if (is_null($title)) {
        $title = "";
    }
    if (is_null($keywords)) {
        $keywords = "";
    }
    $hostarray = uniqueArray(
        explode(
            " ",
            preg_replace(
                "/[^[:alnum:]-]+/i", " ", mb_strtolower($host ?? '', 'utf-8')
            )
        )
    );
    $patharray = uniqueArray(
        explode(
            " ",
            preg_replace(
                "/[^[:alnum:]-]+/i", " ", mb_strtolower($path ?? '', 'utf-8')
            )
        )
    );
    $titlearray = uniqueArray(
        explode(
            " ",
            preg_replace(
                "/[^[:alnum:]-]+/i", " ", mb_strtolower($title ?? '', 'utf-8')
            )
        )
    );
    $keywordsarray = uniqueArray(
        explode(
            " ",
            preg_replace(
                "/[^[:alnum:]-]+/i", " ", mb_strtolower($keywords ?? '', 'utf-8')
            )
        )
    );
    $path_depth = countSubStrs($path, "/");

    foreach ($wordarray as $wid => $word) {
        $word_in_path = 0;
        $word_in_domain = 0;
        $word_in_title = 0;
        $meta_keyword = 0;
        if ($index_host == 1) {
            foreach ($patharray as $id => $path) {
                if ($path[1] == $word[1]) {
                    $word_in_path = 1;
                    break;
                }
            }
            reset($patharray);
            foreach ($hostarray as $id => $host) {
                if ($host[1] == $word[1]) {
                    $word_in_domain = 1;
                    break;
                }
            }
            reset($hostarray);
        }
        if ($index_meta_keywords == 1) {
            foreach ($keywordsarray as $id => $keyword) {
                if ($keyword[1] == $word[1]) {
                    $meta_keyword = 1;
                    break;
                }
            }
            reset($keywordsarray);
        }
        foreach ($titlearray as $id => $tit) {
            if ($tit[1] == $word[1]) {
                $word_in_title = 1;
                break;
            }
        }
        reset($titlearray);
        $wordarray[$wid][2] = (int) (calcWeight(
            $wordarray[$wid][2],
            $word_in_title, $word_in_domain, $word_in_path, $path_depth,
            $meta_keyword
        ));
    }
    reset($wordarray);
    return $wordarray;
}


/**
 * Check for duplicates
 *
 * @param string $md5sum MD5SUM to check
 *
 * @return bool Returns true or false
 */
function isDuplicateMD5($md5sum)
{
    global $mysql_table_prefix, $db;

    $stmt = $db->prepare(
        "SELECT link_id FROM ".$mysql_table_prefix
        ."links WHERE md5sum = ? "
    );
    if ($stmt) {
        $stmt->bind_param("s", $md5sum);
        $stmt->execute() or die("Execution failed: ".$stmt->error);
        $stmt->store_result();
        $num_rows = $stmt->num_rows;
        $stmt->free_result();
        $stmt->close();
    } else {
        trigger_error('Statement failed : '.$db->error, E_USER_ERROR);
    }
    if ($num_rows > 0) {
        return true;
    }
    return false;
}


/**
 * Check for url patterns to include or exclude
 *
 * @param string $link    The url being checked
 * @param string $inc     Patterns to include
 * @param string $not_inc Patterns to exclude
 *
 * @return bool $include Url should or should not be included
 */
function checkInclude($link, $inc, $not_inc)
{
    $url_inc = Array ();
    $url_not_inc = Array ();
    if ($inc != "") {
        $url_inc = explode("\n", $inc);
    }
    if ($not_inc != "") {
        $url_not_inc = explode("\n", $not_inc);
    }
    $oklinks = Array ();

    $include = true;
    foreach ($url_not_inc as $str) {
        $str = trim($str);
        if ($str != "") {
            if (mb_substr($str ?? '', 0, 1, 'utf-8') == '*') {
                if (preg_match(
                    mb_substr($str ?? '', 1, null, 'utf-8'), $link ?? ''
                )
                ) {
                    $include = false;
                    break;
                }
            } else {
                if (!(mb_strpos($link, $str, 0, 'utf-8') === false)) {
                    $include = false;
                       break;
                }
            }
        }
    }
    if ($include && $inc != "") {
        $include = false;
        foreach ($url_inc as $str) {
            $str = trim($str);
            if ($str != "") {
                if (mb_substr($str ?? '', 0, 1, 'utf-8') == '*') {
                    if (preg_match(
                        mb_substr($str ?? '', 1, null, 'utf-8'), $link ?? ''
                    )
                    ) {
                        $include = true;
                        break;
                    }
                } else {
                    if (mb_strpos($link, $str, 0, 'utf-8') !== false) {
                        $include = true;
                        break;
                    }
                }
            }
        }
    }
    return $include;
}


/**
 * Check if a url should be removed
 *
 * @param string $url Url to be checked
 *
 * @return void
 */
function checkForRemoval($url)
{
    global $mysql_table_prefix, $command_line, $db;

    $stmt = $db->prepare(
        "SELECT link_id, visible FROM ".$mysql_table_prefix
        ."links WHERE url = ? "
    );
    if ($stmt) {
        $stmt->bind_param("s", $url);
        $stmt->execute() or die("Execution failed: ".$stmt->error);
        $result = $stmt->get_result();
        $stmt->close();
    } else {
        trigger_error('Statement failed : '.$db->error, E_USER_ERROR);
    }
    if ($result->num_rows > 0) {
        $row = $result->fetch_array(MYSQLI_NUM);
        $link_id = $row[0];
        $visible = $row[1];
        if ($visible > 0) {
            $visible --;
            $ustmt = $db->prepare(
                "UPDATE ".$mysql_table_prefix
                ."links SET visible = ? WHERE link_id = ? "
            );
            if ($ustmt) {
                $ustmt->bind_param("ii", $visible, $link_id);
                if (!$ustmt->execute()) {
                    // Less stringent than die due to possible errors in
                    // pages being indexed
                    trigger_error(
                        "Execution failed: "
                        .$db->error, E_USER_WARNING
                    );
                }
                $ustmt->close();
            } else {
                trigger_error('Statement failed : '.$db->error, E_USER_ERROR);
            }
        } else {
            $dstmt = $db->prepare(
                "DELETE FROM ".$mysql_table_prefix
                ."links WHERE link_id = ? "
            );
            if ($dstmt) {
                $dstmt->bind_param("i", $link_id);
                $dstmt->execute() or die("Execution failed: ".$dstmt->error);
                $dstmt->close();
            } else {
                trigger_error('Statement failed : '.$db->error, E_USER_ERROR);
            }
            for ($i=0;$i<=15; $i++) {
                $char = dechex($i);
                $dstmt2 = $db->prepare(
                    "DELETE FROM ".$mysql_table_prefix
                    ."link_keyword".$char." WHERE link_id = ? "
                );
                if ($dstmt2) {
                    $dstmt2->bind_param("i", $link_id);
                    $dstmt2->execute() or die("Execution failed: ".$dstmt2->error);
                    $dstmt2->close();
                } else {
                    trigger_error(
                        'Statement failed : '
                        .$db->error, E_USER_ERROR
                    );
                }
            }
            printStandardReport('pageRemoved', $command_line);
        }
    }
}


/**
 * Convert certain url characters
 *
 * @param string $url Url to be examined/changed
 *
 * @return string $url Updated url
 */
function convertUrl($url)
{
    $url = str_replace("&amp;", "&", $url);
    $url = str_replace(" ", "%20", $url);
    return $url;
}


/**
 * Extract text from pdf, doc, xls, and ppt files
 *
 * @param array  $contents    File contents
 * @param string $source_type Is the file pdf, doc, xls,
 *                            ppt, docx or odt
 *
 * @return array $result Extracted text array
 */
function extractText($contents, $source_type)
{
    global $tmp_dir, $pdftotext_path, $catdoc_path, $xls2csv_path,
    $catppt_path, $pandoc_path;

    $temp_file = "tmp_file";
    $filename = $tmp_dir."/".$temp_file;
    if (!$handle = fopen($filename, 'w')) {
        die("Cannot open file $filename");
    }

    if (fwrite($handle, $contents) === false) {
        die("Cannot write to file $filename");
    }
    
    fclose($handle);
    if ($source_type == 'pdf') {
        $command = $pdftotext_path." -enc UTF-8 ".escapeshellarg($filename)." -";
        $a = exec(escapeshellcmd($command), $result, $retval);
    } else if ($source_type == 'doc') {
        $command = $catdoc_path." ".escapeshellarg($filename);
        $a = exec(escapeshellcmd($command), $result, $retval);
    } else if ($source_type == 'xls') {
        $command = $xls2csv_path." ".escapeshellarg($filename);
        $a = exec(escapeshellcmd($command), $result, $retval);
    } else if ($source_type == 'ppt') {
        $command = $catppt_path." ".escapeshellarg($filename);
        $a = exec(escapeshellcmd($command), $result, $retval);
    } else if ($source_type == 'docx' || $source_type == 'odt'
    ) {
        $command = $pandoc_path." -f ".$source_type." -t plain "
        .escapeshellarg($filename);
        $a = exec(escapeshellcmd($command), $result, $retval);
    }

    unlink($filename);
    return implode(' ', $result);

}


/**
 * Extract text from pdf, doc, xls, and ppt files
 * Windows specific version of extractText, modification adapted
 * from a 2015 mod by doogle
 *
 * @param array  $contents    File contents
 * @param string $source_type Is the file pdf, doc, xls,
 *                            ppt, docx or odt
 *
 * @return array $result Extracted text array
 */
function winExtractText($contents, $source_type)
{
    global $tmp_dir, $pdftotext_path, $catdoc_path, $xls2csv_path,
    $catppt_path, $pandoc_path;

    $ext = ".".$source_type;
    $temp_file = "tmp_file".$ext;
    $filename = $tmp_dir."\\".$temp_file;
    if (!$handle = fopen($filename, 'w')) {
        die("Cannot open file $filename");
    }

    if (fwrite($handle, $contents) === false) {
        die("Cannot write to file $filename");
    }

    fclose($handle);
    if ($source_type == 'pdf') {
        $command = $pdftotext_path." -enc UTF-8 ".escapeshellarg($filename)." -";
        $a = exec(escapeshellcmd($command), $result, $retval);
    } else if ($source_type == 'doc') {
        $command = $catdoc_path." ".escapeshellarg($filename);
        $a = exec(escapeshellcmd($command), $result, $retval);
    } else if ($source_type == 'xls') {
        $command = $xls2csv_path." ".escapeshellarg($filename);
        $a = exec(escapeshellcmd($command), $result, $retval);
    } else if ($source_type == 'ppt') {
        $command = $catppt_path." ".escapeshellarg($filename);
        $a = exec(escapeshellcmd($command), $result, $retval);
    } else if ($source_type == 'docx' || $source_type == 'odt'
    ) {
        $command = $pandoc_path." -f ".$source_type." -t plain "
        .escapeshellarg($filename);
        $a = exec(escapeshellcmd($command), $result, $retval);
    }

    unlink($filename);
    return implode(' ', $result);

} 


/**
 * Function to calculate the weight of pages
 *
 * @param int $words_in_page  Words on the page
 * @param int $word_in_title  Words in the title
 * @param int $word_in_domain Words in the domain name
 * @param int $word_in_path   Words in the path
 * @param int $path_depth     Depth of the path
 * @param int $meta_keyword   Words in meta tags
 *
 * @return int $weight Calculated word weight
 */
function calcWeight($words_in_page, $word_in_title, $word_in_domain,
    $word_in_path, $path_depth, $meta_keyword
) {

    global $title_weight, $domain_weight, $path_weight,$meta_weight;

    $weight = ($words_in_page + $word_in_title * $title_weight +
    $word_in_domain * $domain_weight +
    $word_in_path * $path_weight + $meta_keyword * $meta_weight)
      * 10 / (0.8 +0.2*$path_depth);

    return $weight;
}


/**
 * Remove Session id
 *
 * @param string $url Url whoses session is to be removed
 *
 * @return string $url Url with session info removed
 */
function removeSessId($url)
{
    if (is_null($url)) {
        $url = "";
    }
    return preg_replace(
        "/(\?|&)(PHPSESSID|JSESSIONID|ASPSESSIONID|sid)=[0-9a-zA-Z]+$/", "", $url
    );
}


/**
 * Function to determine the charset from headers or html
 *
 * @param string $url Url from which to read header to determin charset
 *
 * @return string $charset The charset, if found, else empty
 */
function readHeaders($url)
{
    $headers = get_headers($url, 1);
    if (!$headers) {
        // get_headers failed - retry
        $headers = get_headers($url, 1);
        if (!$headers) {
            // Failed again - give up
            $headers = array();
        }
    }
    $regs = array();
    $charset2 = ""; 
    if (array_key_exists("Content-Type", $headers)) {
        if (is_array($headers['Content-Type'])) {
            $ref = strpos($headers['Content-Type'][0], "charset");
            if ($ref !== false) {
                $set = substr($headers['Content-Type'][0], $ref);
                $ref2 = strpos($set, "=");
                if ($ref2 !== false) {
                    $set2 = substr($set, $ref2+1);
                    $charset = trim($set2, ":; \t\n\r\0\x0B");
                } else {
                    $charset = "";
                }
            } else {
                $charset = "";
            }
        } else {
            $ref = strpos($headers['Content-Type'], "charset");
            if ($ref !== false) {
                $set = substr($headers['Content-Type'], $ref);
                $ref2 = strpos($set, "=");
                if ($ref2 !== false) {
                    $set2 = substr($set, $ref2+1);
                    $charset = trim($set2, ":; \t\n\r\0\x0B");
                } else {
                    $charset = "";
                }
            } else {
                $charset = "";
            }
        }
    } else {
        $charset = "";
    }
    
    $file = @file_get_contents($url);
    if (preg_match(
        '@<meta .*charset *= *([_a-zA-Z0-9 \'"-\.]*) *.*>@i',
        $file ?? '', $regs
    )
    ) {
        $charset2 = preg_replace('/charset=/', '', $regs[1] ?? '');
        $charset2 = preg_replace('/["\']/', '', $charset2 ?? ''); 
    }
    
    // If character set is defined on the page, this overrides
    // the character set defined in the headers
    if ($charset2 != "" && $charset != $charset2) {
        $charset = $charset2;
    }

    // The purpose of this is to protect execution from
    // INSANELY large pages or files
    if ($charset == "" && strlen($file) > 100000) {
        return $charset;
    }

    // If charset is STILL not defined, try to determine charset
    // from the text
    if ($charset == "" && function_exists('mb_detect_encoding')) {
        $file = strip_tags($file);
        $encoding = mb_detect_encoding($file);
        if (isset($encoding)) {
            $charset = $encoding;
        }
    }
    
    // After all that, sometimes we STILL don't know the character set!
    // Set to ISO-8859-1 (latin-1) as a default
    if ($charset == "") {
        $charset = "ISO-8859-1";
    }
    
    // Do this to eliminate ANY confusion later due to case!
    if ($charset == "utf-8") {
        $charset = "UTF-8";
    }
    
    return $charset;

}

/**
 * Function to determine if the url should be allowed
 *
 * @param string $url      Url being considered
 * @param string $deny_str The deny string
 * @param array  $allows   Array of allow string
 *
 * @return bool  true/false   Allow or deny
 */
function checkAllow($url, $deny_str, $allows)
{
    $urlparts = parse_url($url);

    foreach ($allows as $allowurl) {
        $allowurl = trim($allowurl);
        $allowurl_parts = parse_url($allowurl);
        if ($allowurl_parts['scheme'] == '') {
            $check_allow = $urlparts['host'] . $allowurl;
        } else {
            $check_allow = $allowurl;
        }
        // A wildcard (*) in in check_allow is open to too much vagueness
        // Perhaps future insight may help!
        if (strpos($check_allow, "*")) {
            continue;
        }

        if (mb_strpos($url, $check_allow, 0, 'utf-8')) {
            // A specific allow overrides a wildcard
            if (strpos($deny_str, "*")) {
                return true;
            }
            // No allow wildcard
            if (mb_strlen($check_allow, 'utf-8') > mb_strlen($deny_str, 'utf-8')) {
                return true;
            } else {
                continue;
            }
            continue;
        } else {
            continue;
        }
        return false;
    }

    return false;
}


/**
 * Function to find language code
 *
 * @param string $file The file to scan
 *
 * @return void
 */
function getPageLanguage($file)
{
    global $pglang, $com_txt_langs;

    // Initialize page language to null
    $pglang = "";
    // Check if a language is specified in the <html> tag
    if (preg_match('@<html.*?lang=["\'](.*?)["\'].*?>@si', $file, $regs)) {
        $pglang = $regs[1];
    }
    // If code has a '-' (except 'zh') truncate to first two characters
    if (preg_match("@.[a-y]@i", $pglang)) {
        $pglang = strtok($pglang, '-');
    }
    // If the code is 'zh', 'cmn', or 'zh-Hans' we will assume Simplified Chinese
    if ($pglang === "zh" || $pglang === "zh-Hans" || $pglang === "cmn") {
        $pglang = "zh-cn";
    }
    // If the code is 'zh-Hant' we will assume Traditional Chinese
    if ($pglang === "zh-Hant") {
        $pglang = "zh-tw";
    }
    // Check if language code is in our list of languages
    if (!is_null($pglang)) {
        $found = 0;
        foreach ($com_txt_langs as $key => $data) {
            if ($pglang ===  $data[0]) {
                $found = 1;
                break;
            }
        }
        // If the code is not in our list, set it back to null
        if ($found == 0) {
            $pglang = "";
        }
    }
}
